package org.openxma.dsl.generator.impl

import com.google.common.collect.Lists
import java.util.List
import javax.inject.Inject
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.xtext.generator.IFileSystemAccess
import org.eclipse.xtext.generator.IGenerator
import org.openxma.dsl.core.model.DataTypeAndTypeParameter
import org.openxma.dsl.core.model.Model
import org.openxma.dsl.core.model.SimpleType
import org.openxma.dsl.dom.model.AliasedExpression
import org.openxma.dsl.dom.model.Attribute
import org.openxma.dsl.dom.model.AttributeGroup
import org.openxma.dsl.dom.model.CallInputParameter
import org.openxma.dsl.dom.model.CallableStatement
import org.openxma.dsl.dom.model.ComplexType
import org.openxma.dsl.dom.model.Dao
import org.openxma.dsl.dom.model.DataBaseConstraint
import org.openxma.dsl.dom.model.DataBaseConstraintType
import org.openxma.dsl.dom.model.DataView
import org.openxma.dsl.dom.model.DeleteStatement
import org.openxma.dsl.dom.model.Entity
import org.openxma.dsl.dom.model.Join
import org.openxma.dsl.dom.model.Operation
import org.openxma.dsl.dom.model.Parameter
import org.openxma.dsl.dom.model.PropertyValue
import org.openxma.dsl.dom.model.QueryOperation
import org.openxma.dsl.dom.model.QueryParameter
import org.openxma.dsl.dom.model.QueryParameterValue
import org.openxma.dsl.dom.model.SelectProperties
import org.openxma.dsl.dom.model.UpdateStatement
import org.openxma.dsl.generator.lib.Index

import static org.openxma.dsl.generator.GeneratorConstants.*
import static org.openxma.dsl.generator.helper.GeneratorConfigurationExtension.*

import static extension org.openxma.dsl.common.Strings.*
import static extension org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.*
import static extension org.openxma.dsl.generator.helper.SimpleTypeExtension.*
import static extension org.openxma.dsl.generator.helper.ValidatorExtension.*
import static extension org.openxma.dsl.generator.lib.ImportAdapter.*
import static extension org.openxma.dsl.generator.lib.IterableExtensions2.*
import static extension org.openxma.dsl.generator.helper.NameExtension.*;


class RepositoryGenerator implements IGenerator {
	@Inject
	extension Names
	@Inject
	extension Files
	@Inject
	extension Features

	override doGenerate(Resource resource, IFileSystemAccess fsa) {
		val model = resource.getEObject("/") as Model
		model.elements.filter(typeof(Dao)).forEach(
			[	fsa.generateFile(qualifiedGenInterfaceName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenInterface)
				fsa.generateFile(qualifiedInterfaceName.toJavaFilePath, OUTLET_JAVA, toInterface)
				fsa.generateFile(qualifiedGenClassName.toJavaFilePath, OUTLET_GENERATED_JAVA, toGenClass)
				fsa.generateFile(qualifiedClassName.toJavaFilePath, OUTLET_JAVA, toClass)
			]
		)
	}

	def toClass(Dao it) {
		'''
		package classPackageName;

		import org.springframework.stereotype.Repository;
		addImportsHere
		
		@Repository("toVariableName()")
		public class getClassName() extends getGenClassName() {
		FOR operation : operations.filter(e|e.expression==null)
			operation.daoOperationSignature("public") {
				throw new IllegalStateException("not implemented");
			}
		ENDFOR
		}
		'''
	}
	
	def toGenInterface(Dao it) {
		val notNullFeatures = entity.requiredReferences.toList
		notNullFeatures.addAll(entity.requiredAttributes.reject([ idOrVersionAttribute]))
		'''
			package packageName;

			import org.openxma.dsl.platform.dao.*;
			addImportsHere
			
			/**
			documentationLines
			 * Represents a Data Access Object, as defined in the
			 * Core J2EE Pattern Catalog (see http://java.sun.com/blueprints/corej2eepatterns/Patterns/DataAccessObject.html),
			 * to abstract and encapsulate all access to entity name.
			 * @see qualifiedName
			 */
			public interface genInterfaceName extends GenericDao<entity.qualifiedName,idClassName>, EntityFactory<entity.qualifiedName> {
				IF !entity.sortOrders.isEmpty
				/**
				 * Sort order constants
				 */
				ENDIF
				FOR sortOrder : entity.sortOrders
				static final SortOrderSpecification sortOrder.constantName = SortOrderSpecification.create("entity.name.sortOrder.name")FOR orderDefinition : sortOrder.attributesIF orderDefinition.asc.asc("orderDefinition.attribute.name")ELSE .desc("orderDefinition.attribute.name")ENDIFENDFOR;
				ENDFOR
				IF !queryOperation.isEmpty
				/**
				 * Query name constants
				 */
				FOR queryOperation :  queryOperation.reject([statement instanceof CallableStatement])
				static final String queryOperation.getQueryConstant = "queryOperation.getQualifiedQueryName";
				ENDFOR
				/**
				 * Stored procedure name constants
				 */
				FOR queryOperation : queryOperation.filter([statement instanceof CallableStatement])
				static final String queryOperation.storedProcedureConstantName = "(queryOperation.statement as CallableStatement).name.toSimpleName";
				ENDFOR
				ENDIF
				createOperationInterface
				FOR constraint : dataBaseConstraints.reject([type == DataBaseConstraintType::PRIMARY])
				constraint.dbConstraintFinderSignature(null)
				ENDFOR
				FOR operation : operations
				operation.daoOperationSignature(null)
				ENDFOR
				FOR queryOperation : queryOperation
				queryOperation.queryOperationSignature(null)
				ENDFOR
			}
		'''
	}

	def createOperationInterface (Dao it) {
		val notNullFeatures = entity.requiredReferences.toList
		notNullFeatures.addAll(entity.requiredAttributes.reject([idOrVersionAttribute]))
		if (primaryKey != null) {
			'''
			/**
			 * Factory method to create an empty <code>entity.name</code> entity with an already assigned identifier.
			 * The advantage of this approach, coupled with the fact that we're using the id for equals,
			 * is that equality test's are immediately possible and therefore independent from the persistence state
			 * of the entity. This in contrast to other solutions where auto generated Keys are used
			 * which must wait until the object is persisted or differentiate between transient and managed states.
			 * @return a new (transient), empty <code>entity.name</code> entity with an already assigned identifier.
			 */
			entity.name create();
			IF !notNullFeatures.empty
			/**
			 * Factory method to create an empty <code>entity.name</code> taking all not null attributes.
			 *
			 FOR feature : notNullFeatures
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return a new (transient ) <code>entity.name</code> entity with the given natural identifier and not null attributes set.
			 */
			entity.name create(FOR field : notNullFeatures SEPARATOR ","field.methodParameterENDFOR);
			/**
			 * Factory method to create and save an empty <code>entity.name</code> taking all not null attributes.
			 *
			 FOR feature : notNullFeatures
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return a new (transient) <code>entity.name</code> entity with the given natural identifier and not null attributes set.
			 */
			entity.name createAndSave(FOR field : notNullFeatures SEPARATOR ","field.methodParameterENDFOR);
			ENDIF
			'''	
		} else {
			val naturalKeyFields = naturalKey.attributes
			val notNullFields = naturalKeyFields.toSet
			'''
			/**
			 * Factory method to create an empty <code>entity.name</code> entity with an already assigned identifier.
			 * The advantage of this approach, coupled with the fact that we're using the id for equals,
			 * is that equality test's are immediately possible and therefore independent from the persistence state
			 * of the entity. This in contrast to other solutions where auto generated Keys are used
			 * which must wait until the object is persisted or differentiate between transient and managed states.
			 FOR feature : naturalKeyFields
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return a new (transient), empty <code>entity.name</code> entity with an already assigned identifier.
			 */
			entity.name create(FOR field : naturalKeyFields SEPARATOR ","field.methodParameterENDFOR);
			IF notNullFields.addAll(notNullFeatures)
			/**
			 * Factory method to create an empty <code>entity.name</code> taking all required (not-null) attributes.
			 *
			 FOR feature : notNullFields
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return a new, transient (i.e. is not associated with a persistence context and has no persistent representation in the database) <code>entity.name</code> entity with the given natural identifier and not null attributes set.
			*/
			entity.name create(FOR field : notNullFields SEPARATOR ","field.methodParameterENDFOR);
			ENDIF
			/**
			 * Factory method to create and save an empty <code>entity.name</code> taking all required (not-null) attributes.
			 *
			  FOR feature : notNullFields
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return a new, transient (i.e. is not associated with a persistence context and has no persistent representation in the database) <code>entity.name</code> entity with the given natural identifier and not null attributes set.
			 */
			entity.name createAndSave(FOR field : notNullFields SEPARATOR ","field.methodParameterENDFOR);
			'''	
		}
	}
	
	def dbConstraintFinderSignature (DataBaseConstraint it, String visibilityModifier) {
		val resolvedAttributeList = resolvedAttributeList
		'''
			IF type == DataBaseConstraintType::NATURAL
			IF visibilityModifier==null
			/**
			 * Return the persistent instance of the entity.name entity class
			 * with the given natural identifier or <code>null</code> if not found.
			 *
			 FOR feature : resolvedAttributeList
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return the persistent entity.name instance, or <code>null</code> if not found
			 */
			ENDIF
			IF visibilityModifier!=nullvisibilityModifier ENDIFentity.name getMethodName()(FOR parameter : resolvedAttributeList SEPARATOR ","parameter.methodParameterENDFOR)IF visibilityModifier==null;ENDIFELSEIF type == DataBaseConstraintType::UNIQUE
			IF visibilityModifier==null
			/**
			 * Return the persistent instance of the entity.name entity class
			 * with the given unique-key or <code>null</code> if not found.
			 *
			 FOR feature : attributes
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @return the persistent entity.name instance, or <code>null</code> if not found
			 */
			ENDIF
			IF visibilityModifier!=nullvisibilityModifier ENDIFentity.name getMethodName()(FOR parameter : attributes SEPARATOR ","parameter.methodParameterENDFOR)IF visibilityModifier==null;ENDIFELSEIF type == DataBaseConstraintType::INDEX
			IF visibilityModifier==null
			/**
			 * Execute a query based on the given index attributes.
			 *
			 FOR feature : resolvedAttributeList
			 feature.methodParameterJavaDoc
			 ENDFOR
			 * @param firstResult the index of the first result object to be retrieved (numbered from 0 or null if ignored)
			 * @param maxResults the maximum number of result objects to retrieved (null or <=0 for no limit)
			 * @return a {@link Set} containing 0 or more persistent getEntity().name instances
			 */
			ENDIF
			IF visibilityModifier!=nullvisibilityModifier ENDIFaddImport("java.util.Set")<getEntity().name> getMethodName()(FOR parameter : resolvedAttributeList SEPARATOR ","parameter.methodParameterENDFOR,Integer firstResult,Integer maxResults")"IF visibilityModifier==null;ENDIFENDIF'''
	}
	
	def daoOperationSignature(Operation it, String visibilityModifier) {
		'''
		IF visibilityModifier!=null
		ELSE
		/**
		 *
		 FOR parameter : parameters
		 parameter.methodParameterJavaDoc
		 ENDFOR
		 IF many
		 * @return a collection containing 0 or more persistent type.name instances
		 ELSE
		 * @return the type?.name.toFirstLower() ?:"" instance, or <code>null</code> if not found
		 ENDIF
		 */
		 ENDIF
		IF visibilityModifier!=nullvisibilityModifier ENDIFtoType name(FOR parameter : parameters SEPARATOR ","parameter.methodParameterENDFOR)IF visibilityModifier==null;ENDIF'''
	}
	
	def queryOperationSignature(QueryOperation it, String visibilityModifier) {
		'''
		IF visibilityModifier!=null
		ELSE
		/**
		documentationLinesIndented
		 *
		 FOR parameter : queryParameters
		 parameter.methodParameterJavaDoc
		 ENDFOR
		 IF many	 
		 * @return a collection containing 0 or more persistent type.name instances
		 ELSE
		 * @return the type?.name.toFirstLower() ?:"" instance, or <code>null</code> if not found
		 ENDIF
		 */
		 ENDIF
		IF visibilityModifier!=nullvisibilityModifier ENDIFtoType name(FOR parameter : queryParameters SEPARATOR ","parameter.methodParameterENDFOR)IF visibilityModifier==nullvisibilityModifier;ENDIF'''
	}

   
	def getStoredProcedureConstantName(QueryOperation it) {
		"SP_" + getQueryConstant
	}

	def getIdClassName(Dao it) {
		if (primaryKey != null) {
			primaryKeyColumn.attribute.dataType.qualifiedName
		} else {
			addImport(entity.key.qualifiedKeyClassName)
		}
	}

	def toInterface(Dao it) {
		'''
		package packageName;
		
		documentation
		public interface name extends getGenInterfaceName() {
		}
		'''
	}

	def toGenClass(Dao it) {
		val notNullFeatures = entity.requiredReferences.toList
		notNullFeatures.addAll(entity.requiredAttributes.reject([idOrVersionAttribute]))
		'''
		package classPackageName;

		import org.openxma.dsl.platform.dao.*;
		import org.openxma.dsl.platform.dao.impl.*;
		import org.openxma.dsl.platform.service.Mapper;
		
		import org.hibernate.Criteria;
		import org.hibernate.Query;
		import org.hibernate.SessionFactory;
		import org.hibernate.criterion.Criterion;
		import org.hibernate.criterion.Expression;
		import org.hibernate.criterion.Restrictions;
		import org.springframework.beans.factory.annotation.Autowired;
		import org.springframework.beans.factory.annotation.Qualifier;
		import org.springframework.jdbc.support.incrementer.DataFieldMaxValueIncrementer;
		import org.springframework.stereotype.Repository;
		import org.springframework.util.Assert;
		addImportsHere
		/**
		 * @see interfaceName
		 */
		@SuppressWarnings("unused")
		public abstract class genClassName extends IF "GenericDaoHibernateImpl".equals(daoBaseClassname.lastToken("."))GenericDaoHibernateImplELSEaddImport(daoBaseClassname)ENDIF<entity.qualifiedName,idClassName> implements qualifiedName {
			
			public genClassName() {
				super(addImport(entity.qualifiedClassName).class);
			}
			
			/**
			 * @param context the context for the object creation
			 * @return a transient instance of entity <code>entity.name</code> with an already assigned identifier.
			 */
			public entity.name createEntity(Object context) {
				IF primaryKey!=null
				return create();
				ELSE
				return createEntityInstance();
				ENDIF
			}
			
			IF primaryKey != null
			public entity.name create() 
			{
				entity.name entity.name.toVariableName() = createEntityInstance();
				IF primaryKeyColumn.attribute.type.getFullyQualifiedName(true).equals("Long")
				entity.name.toVariableName().primaryKeyColumn.attribute.getMutatorMethodName()(Long.valueOf(dataFieldMaxValueIncrementer.nextLongValue()));
				ELSEIF primaryKeyColumn.attribute.type.getFullyQualifiedName(true).equals("String")
				entity.name.toVariableName().primaryKeyColumn.attribute.getMutatorMethodName()(dataFieldMaxValueIncrementer.nextStringValue());
				ELSEIF primaryKeyColumn.attribute.type.getFullyQualifiedName(true).equals("Integer")
				entity.name.toVariableName().primaryKeyColumn.attribute.getMutatorMethodName()(dataFieldMaxValueIncrementer.nextIntValue());
				ELSEIF "org.openxma.dsl.platform.valueobject.ValueObjectString".equals((primaryKeyColumn.attribute.dataType as SimpleType).getJavaType)
				entity.name.toVariableName().primaryKeyColumn.attribute.getMutatorMethodName()(new primaryKeyColumn.attribute.dataType.qualifiedName(dataFieldMaxValueIncrementer.nextStringValue()));
				ELSEIF "org.openxma.dsl.platform.valueobject.ValueObjectLong".equals((primaryKeyColumn.attribute.dataType as SimpleType).getJavaType)
				entity.name.toVariableName().primaryKeyColumn.attribute.getMutatorMethodName()(new primaryKeyColumn.attribute.dataType.qualifiedName(dataFieldMaxValueIncrementer.nextLongValue()));
				ENDIF
				return entity.name.toVariableName();
			}
			IF !notNullFeatures.empty
			
			public entity.name create(FOR field : notNullFeatures SEPARATOR ","field.methodParameterENDFOR) {
				FOR field : notNullFeatures
				Assert.notNull(field.getResolvedName().toFirstLower(), "Parameter 'field.getResolvedName().toFirstLower()' must not be null");
				ENDFOR
				entity.name entity.name.toVariableName() = (entity.name) create();
				FOR field : notNullFeatures
				entity.name.toVariableName().field.getMutatorMethodName()(field.getResolvedName().toFirstLower());
				ENDFOR
				return entity.name.toVariableName();
			}
			
			public entity.name createAndSave(FOR field : notNullFeatures SEPARATOR ","field.methodParameterENDFOR) {
				entity.name entity.name.toVariableName() = create(FOR field : notNullFeatures SEPARATOR ","field.resolvedName.toFirstLowerENDFOR);
				saveOrUpdate(entity.name.toVariableName());
				return entity.name.toVariableName();
			}
			ENDIF
			ELSE
			val naturalKeyFields = naturalKey.attributes
			val notNullFields = naturalKeyFields.toSet
			public entity.name create(FOR field : naturalKeyFields SEPARATOR ","field.methodParameterENDFOR) {
				FOR field : naturalKeyFields
				Assert.notNull(field.toVariableName, "Parameter 'field.resolvedName.toFirstLower()' must not be null");
				ENDFOR
				entity.className entity.name.toVariableName() = new entity.className();
				FOR field : naturalKeyFields
				entity.name.toVariableName().field.getMutatorMethodName()(field.resolvedName.toFirstLower());
				ENDFOR
				return entity.name.toVariableName();
			}
			IF notNullFields.addAll(notNullFeatures)
			
			public entity.name create(FOR field : notNullFields SEPARATOR ","field.methodParameterENDFOR) {
				entity.name entity.name.toVariableName() = (entity.name) create(FOR field : naturalKeyFields SEPARATOR ","field.nameENDFOR);
				FOR field : notNullFeatures
				Assert.notNull(field.getResolvedName().toFirstLower(), "Parameter 'field.getResolvedName().toFirstLower()' must not be null");
				ENDFOR
				FOR field : notNullFeatures
				entity.name.toVariableName().field.getMutatorMethodName()(field.getResolvedName().toFirstLower());
				ENDFOR
				return entity.name.toVariableName();	
			}
			ENDIF
			
			public entity.name createAndSave(FOR field : notNullFields SEPARATOR ","field.methodParameterENDFOR) {
				entity.name entity.name.toVariableName() = create(FOR field : notNullFields SEPARATOR ','field.getResolvedName().toFirstLower()ENDFOR);
				saveOrUpdate(entity.name.toVariableName());
				return entity.name.toVariableName();
			}
			ENDIF
			FOR constraint : dataBaseConstraints.reject([type == DataBaseConstraintType::PRIMARY])
			
			constraint.toFinderQueryCall
			ENDFOR
			FOR operation : operations.filter(e|e.expression!=null)
			
			operation.toNamedQueryCall
			ENDFOR
			FOR queryOperation : queryOperation
			IF queryOperation.statement instanceof CallableStatement
			
			queryOperation.toSimpleJdbcCall
			ELSE
			
			queryOperation.toNamedQueryCall
			ENDIF
			ENDFOR
			
			protected Mapper getNamedQueryMapper(String queryName) {
				return new nameQueryMapper();
			}
			
			protected class nameQueryMapper extends NamedQueryMapper {
				
			@SuppressWarnings("unchecked")
			public <E> E mapOne(Object sourceObject, E targetObject, Object context) {
				FOR qlOperation : queryOperation.filter(e|e.many && e.type instanceof DataView && e.statement instanceof SelectProperties)
				if ("qlOperation.getQualifiedQueryName".equals(getQueryName(context))) {
					return (E) mapqlOperation.name.toFirstUpper()Tuple((Object[]) sourceObject, 1);
				}
				ENDFOR
				return super.mapOne(sourceObject, targetObject);
			}
		}
		}
		'''
	} 
	
	def toFinderQueryCall(DataBaseConstraint it) {
		val resolvedAttributes = resolvedAttributeList
		'''
		dbConstraintFinderSignature("public") {
			IF DataBaseConstraintType::NATURAL == type || DataBaseConstraintType::INDEX == type
			val index = new Index
			FOR attribute : resolvedAttributes
			IF index.firstIteration
			Criterion criterion = equal("getFinderQueryParameterName(attribute)", attribute.getResolvedName(), !attribute.required);
			ELSE
			criterion = and(criterion,"getFinderQueryParameterName(attribute)", attribute.getResolvedName(), !attribute.required);
			ENDIF
			index.increment
			ENDFOR
			ELSE
			val index = new Index
			FOR field : attributes
			IF index.firstIteration
			Criterion criterion = equal("getFinderQueryPropertyName(field)", field.getResolvedName(), !field.required);
			ELSE
			criterion = and(criterion,"getFinderQueryPropertyName(field)", field.getResolvedName(), !field.required);
			ENDIF
			index.increment
			ENDFOR
			ENDIF
			Criteria criteria = createCriteria(criterion);
			IF DataBaseConstraintType::NATURAL == type || DataBaseConstraintType::INDEX == type
			FOR alias : resolvedAttributes.filter(e|e.getEntity()!=it.getEntity() && e.getEntity()?.repository!=null && e.getEntity()!=it.getEntity().superType).map(e|e.getFQNProperty()).toSet()
			criteria.createAlias("alias.toFirstLower()","alias.toFirstLower()");
			ENDFOR
			ELSE
			FOR alias : getAliasNames()
			criteria.createAlias("alias","alias");
			ENDFOR
			ENDIF
			IF DataBaseConstraintType::NATURAL == type || DataBaseConstraintType::UNIQUE == type
			return (entity.qualifiedName) unique(criteria);
			ELSE
			applyPagingParameters(criteria,firstResult,maxResults);
			return new addImport("java.util.LinkedHashSet")<getEntity().qualifiedName>(list(criteria));
			ENDIF
		}
		'''
	}
	
	def getFinderQueryParameterName(DataBaseConstraint it, Attribute attribute) {
		if (attribute.getEntity()!=getEntity() && attribute.getEntity()?.repository!=null && attribute.getEntity()!=getEntity().superType) {
			attribute.getResolvedParameterName()
		} else {
			attribute.name
		}
	}
	
	def getFinderQueryPropertyName(DataBaseConstraint it, Attribute field) {
		if (field.getFQNProperty !=null && field.getFQNProperty!=getEntity().name) {
			field.getResolvedParameterName()  
		} else {
			field.name	
		}
	}
	
	
	def private getResolvedParameterName(Attribute it) {
		if (getFQNProperty() != null) {
			getFQNProperty().toString().toFirstLower() + "." + name.toFirstLower()
		} else {
			name
		}
	}
	
	
	
	def toSimpleJdbcCall(QueryOperation it) {
		val callableStatement = statement as CallableStatement
		'''
		queryOperationSignature("public") {
			SimpleJdbcCall jdbcCall = preparename.toFirstUpper()JdbcCall(FOR parameter : queryParameters SEPARATOR ","parameter.nameENDFOR);
			Map<String, Object> _resultMap = executename.toFirstUpper()(jdbcCallIF !queryParameters.isEmpty,ENDIFFOR parameter : queryParameters SEPARATOR ","parameter.nameENDFOR);
			IF type!=null
			return mapname.toFirstUpper()Result(_resultMapIF !queryParameters.isEmpty,ENDIFFOR parameter : queryParameters SEPARATOR ","parameter.nameENDFOR);
			ENDIF
		}
		
		protected SimpleJdbcCall preparename.toFirstUpper()JdbcCall(FOR parameter : queryParameters SEPARATOR ","parameter.methodParameterENDFOR) {
			addImport("org.springframework.jdbc.core.simple.SimpleJdbcCall") jdbcCall = createJdbcCall(getStoredProcedureConstantName()IF callableStatement.name.contains("."),"callableStatement.name.toPackageName()"ENDIF);
			IF callableStatement.functionCall
			jdbcCall.setFunction(true);
			ENDIF
			IF many
			IF type instanceof ComplexType
			jdbcCall.addDeclaredRowMapper("_resultSet", new addImport("org.springframework.jdbc.core.simple.ParameterizedRowMapper")<type.name>() {
				public type.name mapRow(addImport("java.sql.ResultSet") rs, int rowNum) throws addImport("java.sql.SQLException") {
					type.name _type.name.toFirstLower() = new type.name();
					FOR outParameter : callableStatement.outParameter.filter(e|e.attribute!=null)
					_type.name.toFirstLower().outParameter.attribute.getMutatorMethodName()(addImport("org.openxma.dsl.platform.jdbc.JdbcUtils2").getResultSetValue(rs,"outParameter.name",outParameter.attribute.type.qualifiedName.class));
					ENDFOR
					return _type.name.toFirstLower();
				}
			});
			ELSE
			jdbcCall.addDeclaredRowMapper("_resultSet", new addImport("org.springframework.jdbc.core.simple.ParameterizedRowMapper")<type.name>() {
				public type.name mapRow(addImport("java.sql.ResultSet") rs, int rowNum) throws addImport("java.sql.SQLException") {
					return addImport("org.openxma.dsl.platform.jdbc.JdbcUtils2").getResultSetValue(rs,"callableStatement.outParameter.head.name",type.qualifiedName.class);
				}
			});
			ENDIF
			ELSEIF type instanceof ComplexType
			FOR outParameter : callableStatement.outParameter.filter(e|e.attribute.many && !(e.attribute.dataType instanceof ComplexType))
			jdbcCall.addDeclaredRowMapper("_outParameter.attribute.name_resultSet", new addImport("org.springframework.jdbc.core.simple.ParameterizedRowMapper")<outParameter.attribute.type.qualifiedName>() {
				public outParameter.attribute.type.qualifiedName mapRow(addImport("java.sql.ResultSet") rs, int rowNum) throws addImport("java.sql.SQLException") {
					return addImport("org.openxma.dsl.platform.jdbc.JdbcUtils2").getResultSetValue(rs,"outParameter.name",outParameter.attribute.type.qualifiedName.class);
				}
			});
			ENDFOR
			val List<Attribute> outParameterAttributes = callableStatement.outParameter.map(p|p.attribute) 
			FOR complexTypeAttribute : outParameterAttributes.filter(e|e.many && e.dataType instanceof ComplexType).toSet
			val index = new Index
			FOR outParameter : callableStatement.outParameter.filter(e|e.attribute==complexTypeAttribute)
			IF index.firstIteration
			jdbcCall.addDeclaredRowMapper("_outParameter.attribute.name_resultSet", new addImport("org.springframework.jdbc.core.simple.ParameterizedRowMapper")<complexTypeAttribute.type.qualifiedName>() {
				public outParameter.attribute.type.qualifiedName mapRow(addImport("java.sql.ResultSet") rs, int rowNum) throws addImport("java.sql.SQLException") {
					complexTypeAttribute.type.qualifiedName _complexTypeAttribute.name.toFirstLower() = new complexTypeAttribute.type.qualifiedName();	   
			ENDIF
					_complexTypeAttribute.name.toFirstLower().outParameter.nestedAttribute.getMutatorMethodName()(addImport("org.openxma.dsl.platform.jdbc.JdbcUtils2").getResultSetValue(rs,"outParameter.name",outParameter.nestedAttribute.type.qualifiedName.class));	 
			index.increment
			ENDFOR
					return _complexTypeAttribute.name.toFirstLower();
				}
			});
			ENDFOR
			ENDIF
			return jdbcCall;
		}
		
		protected Map<String, Object> executename.toFirstUpper()(SimpleJdbcCall jdbcCallIF !queryParameters.isEmpty,ENDIFFOR parameter : queryParameters SEPARATOR ","parameter.methodParameterENDFOR) {
			addImport("java.util.Map")<String,Object> parametersMap = new addImport("java.util.HashMap")<String,Object>();
			FOR inParameter : callableStatement.inParameter
			parametersMap.put("inParameter.getParameterName",IF inParameter.attribute!=nullinParameter.parameter.getName()==null ? null:inParameter.parameter.getName().inParameter.attribute.getAccessorMethodName()()ELSEinParameter.parameter.getName()ENDIF);
			ENDFOR
			return executeJdbcCall(jdbcCall,parametersMap);
		}
		IF type!=null

		protected toType() mapname.toFirstUpper()Result(Map<String, Object> _resultMapIF !queryParameters.isEmpty,ENDIFFOR parameter : queryParameters SEPARATOR ","parameter.methodParameterENDFOR) {
			IF many
			@SuppressWarnings("unchecked")
			toType _resultSet = (toType) _resultMap.get("_resultSet");
			return _resultSet;
			ELSEIF type instanceof ComplexType
			type.name _type.name.toFirstLower() = new type.name();
			FOR outParameter : callableStatement.outParameter.filter(e|!(e.attribute.dataType instanceof ComplexType))
			IF outParameter.attribute.many
			@SuppressWarnings("unchecked")
			addImport("java.util.List")<outParameter.attribute.type.qualifiedName> _outParameter.attribute.nameResultSet = (List<outParameter.attribute.type.qualifiedName>) _resultMap.get("_outParameter.attribute.name_resultSet");
			_type.name.toFirstLower().outParameter.attribute.getMutatorMethodName()(_outParameter.attribute.nameResultSet);
			ELSEIF !outParameter.attribute.many
			_type.name.toFirstLower().outParameter.attribute.getMutatorMethodName()((outParameter.attribute.type.qualifiedName)_resultMap.get("outParameter.name"));
			ENDIF
			ENDFOR
			FOR complexTypeAttribute : callableStatement.outParameter.map(a|a.attribute).<Attribute>filter(e|e.many && e.dataType instanceof ComplexType).toSet()
			@SuppressWarnings("unchecked")
			addImport("java.util.List")<complexTypeAttribute.type.qualifiedName> _complexTypeAttribute.nameResultSet = (List<complexTypeAttribute.type.qualifiedName>) _resultMap.get("_complexTypeAttribute.name_resultSet");
			_type.name.toFirstLower().complexTypeAttribute.getMutatorMethodName()(_complexTypeAttribute.nameResultSet);
			ENDFOR
			FOR complexTypeAttribute : callableStatement.outParameter.map(a|a.attribute).<Attribute>filter(e|!e.many && e.dataType instanceof ComplexType).toSet()
			val index = new Index
			FOR outParameter : callableStatement.outParameter.filter(e|e.attribute==complexTypeAttribute && e.nestedAttribute!=null)
			IF index.firstIteration
			complexTypeAttribute.dataType.qualifiedName _complexTypeAttribute.name = new complexTypeAttribute.dataType.qualifiedName();
			_type.name.toFirstLower().complexTypeAttribute.getMutatorMethodName()(_complexTypeAttribute.name);
			ENDIF
			_complexTypeAttribute.name.outParameter.nestedAttribute.getMutatorMethodName()((outParameter.nestedAttribute.type.qualifiedName)_resultMap.get("outParameter.name"));
			index.increment
			ENDFOR
			ENDFOR
			return _type.name.toFirstLower();
			ELSE
			return (toType) IF callableStatement.functionCall_resultMap.get("return")ELSE_resultMap.get("callableStatement.outParameter.head.name")ENDIF;
			ENDIF	
		}
		ENDIF
		'''
	}
	
	def private getParameterName(CallInputParameter it) {
		if (name!=null) {
			name	
		} else if (attribute!=null) {
			attribute.name	
		} else {
			parameter?.name
		}
	}
	
	def toNamedQueryCall(Operation it) {
		'''
		daoOperationSignature("public") {
			FOR parameter : parameters.filter(e|e.name!="firstResult" && e.name!="maxResults")
			Assert.notNull(parameter.toVariableName, "Parameter 'parameter.toVariableName' must not be null");
			ENDFOR
			Query namedQuery = this.sessionFactory.getCurrentSession().getNamedQuery("(eContainer as Dao).entity.name.name.toFirstUpper()");
			IF expression.contains("?")
			val iter = new Index
			FOR parameter : parameters
			namedQuery.setParameter(iter.counter0, parameter.name);
			iter.increment
			ENDFOR
			ELSE
			FOR parameter : parameters
			IF "firstResult".equals(parameter.name)
			if (firstResult!=null) {
			    namedQuery.setFirstResult(parameter.name);
			}
			ELSEIF "maxResults".equals(parameter.name)
			if (maxResults!=null) {
			    namedQuery.setMaxResults(parameter.name);
			}
			ELSEIF parameter.type instanceof SimpleType && (parameter.type as SimpleType).validatorReference.isDateValidator()
			namedQuery.setDateIF parameter.manyListENDIF("parameter.name", parameter.name);
			ELSEIF parameter.type instanceof SimpleType && (parameter.type as SimpleType).validatorReference.isTimestampValidator()
			namedQuery.setTimestampIF parameter.manyListENDIF("parameter.name", parameter.name);
			ELSEIF parameter.type instanceof Entity && !parameter.many
			namedQuery.setEntity("parameter.name", parameter.name);
			ELSE
			namedQuery.setParameterIF parameter.manyListENDIF("parameter.name", parameter.name);
			ENDIF
			ENDFOR
			ENDIF
			applyname.toFirstUpper()QueryHints(namedQueryIF !parameters.isEmpty,ENDIFFOR parameter : parameters SEPARATOR "," parameter.nameENDFOR);
			return IF manylist(namedQuery)ELSEunique(namedQuery)ENDIF;
		}
		
		protected void applyname.toFirstUpper()QueryHints(Query queryIF !parameters.isEmpty,ENDIFFOR parameter : parameters SEPARATOR ","parameter.methodParameterENDFOR) {
			//apply other hints that influence how a query is executed
		}
		'''	
	}
	
	def toNamedQueryCall(QueryOperation it) {
		'''
		queryOperationSignature("public") {
			String queryName = getQualifiedQueryConstant;
			Query namedQuery = this.sessionFactory.getCurrentSession().getNamedQuery(queryName);
			FOR parameter : queryParameters.reject(["firstResult".equals(name) || "maxResults".equals(name)])
			val value = statement.eAllContents.filter(typeof(QueryParameterValue)).findFirst([value|value.parameter==parameter || value.parameter==parameter.attribute])
			IF !value?.many && value?.type instanceof Entity
			namedQuery.setEntity("parameter.getName()", parameter.getName());
			ELSEIF parameter.type instanceof ComplexType && value?.attribute != null
			FOR attribute : statement.eAllContents.filter(typeof(QueryParameterValue)).filter(e|e.parameter==parameter).map(e|e.attribute).toSet().sortBy(e|e.name)
			namedQuery.attribute.type.getSetParameterMethodName()IF attribute.manyListENDIF("parameter.getName()_attribute.name", parameter.getName().attribute.getAccessorMethodName()());
			ENDFOR
			ELSEIF value != null
			namedQuery.parameter.getSetParameterMethodName()IF parameter.isMany()ListENDIF("parameter.getName()", parameter.getName());
			ENDIF
			ENDFOR
			applyname.toFirstUpper()QueryHints(namedQueryIF !queryParameters.isEmpty,ENDIFFOR parameter : queryParameters SEPARATOR "," parameter.getName()ENDFOR);
			IF statement instanceof UpdateStatement || statement instanceof DeleteStatement
			IF type!=nullreturn ENDIFnamedQuery.executeUpdate();
			ELSEIF type instanceof DataView && statement instanceof SelectProperties
			IF many
			return getNamedQueryMapper(queryName).createAndMapMany(list(namedQuery), type.name.class, addImport("java.util.Collections").singletonMap(QUERY_NAME, queryName));
			ELSE
			Object[] tuple = unique(namedQuery);
			return mapname.toFirstUpper()Tuple(tuple);
			ENDIF
			ELSEIF many
			IF !statement.eAllContents.filter(typeof(Join)).isEmpty
			Collection<type.name> result = list(namedQuery);
			return new addImport("java.util.LinkedHashSet")<type.name>(result);
			ELSE
			return list(namedQuery);
			ENDIF
			ELSEIF type!=null
			return unique(namedQuery);
			ENDIF
		}
		
		protected void applyname.toFirstUpper()QueryHints(Query queryFOR parameter : queryParameters BEFORE "," SEPARATOR ","parameter.toType parameter.nameENDFOR) {
			FOR parameter : queryParameters.filter(typeof(Parameter))
			IF parameter.getName()=="firstResult"
			if (firstResult!=null) {
				query.setFirstResult(parameter.getName());
			}
			ELSEIF parameter.getName()=="maxResults"
			if (maxResults!=null) {
				query.setMaxResults(parameter.getName());
			}
			ENDIF
			ENDFOR
			// override this method and apply other hints that influence how a query is executed
		}
		IF type instanceof DataView && statement instanceof SelectProperties

		val dataView = type as DataView
		val selectProperties = statement as SelectProperties
		protected type.name mapname.toFirstUpper()Tuple(Object[] tupleIF many, int rowNumberENDIF) {
			type.name type.name.toFirstLower() = null;
			if (tuple!=null) {
				type.name.toFirstLower() = new type.name();
				var index = new Index
				FOR expression : selectProperties.properties
				IF expression instanceof AliasedExpression && dataView.allAttributes.findFirst(e|e.name.equals((expression as AliasedExpression).name)) != null
				val attribute = dataView.allAttributes.findFirst(e|e.name.equals((expression as AliasedExpression).name))
				type.name.toFirstLower().attribute.getMutatorMethodName()((attribute.type.qualifiedName)tuple[index.counter0]);
				ELSEIF expression instanceof PropertyValue && dataView.allAttributes.findFirst(e|e.name.equals((expression as PropertyValue).segments.last())) != null
				val attribute = dataView.allAttributes.findFirst(e|e.name.equals((expression as PropertyValue).segments.last()))
				type.name.toFirstLower().attribute.getMutatorMethodName()((attribute.type.qualifiedName)tuple[index.counter0]);
				ENDIF
				index.increment
				ENDFOR
			}
			return type.name.toFirstLower();
		}
		ENDIF
		'''	
	}
	
	def String getQualifiedCompositeIdentifierName(Entity it) { 
		addImport(packageName+"."+key.name)
	}
	
	def dispatch getEntity(Attribute it) {
		eContainer as Entity
	}
	
	def dispatch getEntity(DataBaseConstraint it) {
		(eContainer as Dao).entity
	}
   
	def isMany(QueryParameterValue it) {
		if (attribute != null) {
			return attribute.many 
		}
		val type = parameter
		switch type {
			Parameter : type.many
			Attribute : type.many
			default : false
		}

	}
	
	def getType(QueryParameterValue it) {
		if (attribute != null) {
			return attribute.dataType 
		}
		val type = parameter
		switch type {
			Parameter : type.type
			Attribute : type.dataType
		}
	}
	
	def getSetParameterMethodName(QueryParameter parameter) {
		val type = parameter.type
		switch type {
			SimpleType case type.validatorReference != null && type.validatorReference.dateValidator : "setDate"
			SimpleType case type.validatorReference != null && type.validatorReference.timeStampValidator : "setTimestamp"
			default : "setParameter"
		}   
	}
	
	def getSetParameterMethodName(DataTypeAndTypeParameter it) {
	   if (hasDateValidator) {
		 "setDate"	
	   } else if (hasTimeStampValidator) {
		 "setTimestamp"	
	   } else {
		   "setParameter"
	   }
		
	}
	
	def isMany(QueryParameter it) {
		switch it {
			Parameter : it.many
			default : attribute.many
		}
	}
	
	def getConstantName(AttributeGroup it) {
		name.toUnderScoreString.toUpperCase
	}
	
	def private toSimpleName(String fqName) {
		fqName.lastToken(".")
	}

	def private toPackageName(String fqName) {
		fqName.skipLastToken('.');
	}
	
	def private Iterable<String> getAliasNames(DataBaseConstraint it) {
		if (getProperty("alias") != null) {
			return getProperty("alias") as Iterable<String>
		}
		Lists::newArrayList
	}
	
 }
